package org.processmining.directlyfollowsmodelminer.model;

import gnu.trove.set.TIntSet;

public interface DirectlyFollowsModel {

	/**
	 * 
	 * @param value
	 * @return The activity, or null if it does not exist.
	 */
	public String getActivityOfIndex(int value);

	/**
	 * 
	 * @return The number of activities in the model.
	 */
	public int getNumberOfActivities();

	/**
	 * 
	 * @return The added activities, in the order they were added.
	 */
	public String[] getAllActivities();

	public Iterable<Integer> getActivitiesIndices();

	/**
	 * 
	 * @param activity
	 * @return The indices of the activity.
	 */
	public TIntSet getIndexOfActivity(String activity);

	/**
	 * Inserts an activity.
	 * 
	 * @param activity
	 * @return the index of the activity.
	 */
	public int addActivity(String activity);

	/**
	 * Adds an array of activities. Typically more efficient than adding the
	 * activities one-by-one.
	 * 
	 * @param activities
	 */
	public void addActivities(String[] activities);

	/**
	 * 
	 * @return Whether the model supports empty (epsilon) traces.
	 */
	public boolean isEmptyTraces();

	/**
	 * Set whether the model supports empty (epsilon) traces.
	 * 
	 * @param emptyTraces
	 */
	public void setEmptyTraces(boolean emptyTraces);

	/*
	 * === edges ===
	 */
	/**
	 * Adds an edge. Has no effect if the edge was already present.
	 * 
	 * @param sourceIndex
	 * @param targetIndex
	 */
	public void addEdge(int sourceIndex, int targetIndex);

	/**
	 * Gives an iterable that iterates over all edges; The edges that are
	 * returned are indices. The iterable supports remove() calls, however calls
	 * to this method might invalidate other iterables.
	 * 
	 * @return
	 */
	public Iterable<Long> getEdges();

	/**
	 * Returns whether the graph contains an edge between source and target.
	 * 
	 * @param sourceIndex
	 * @param targetIndex
	 * @return
	 */
	public boolean containsEdge(int sourceIndex, int targetIndex);

	/**
	 * Returns the node the edgeIndex comes from.
	 * 
	 * @param edgeIndex
	 * @return
	 */
	public int getEdgeSource(long edgeIndex);

	/**
	 * Returns the node the edgeIndex points to.
	 * 
	 * @param edgeIndex
	 * @return
	 */
	public int getEdgeTarget(long edgeIndex);

	/*
	 * === start activities ===
	 */
	public boolean hasStartActivities();

	/**
	 * Has no effect if the activity was already a start activity.
	 * 
	 * @param activity
	 */
	public void addStartActivity(int activity);

	public TIntSet getStartActivities();

	public int getNumberOfStartActivities();

	/*
	 * === end activities ===
	 */
	public boolean hasEndActivities();

	/**
	 * Has no effect if the activity was already an end activity.
	 * 
	 * @param activity
	 */
	public void addEndActivity(int activity);

	public TIntSet getEndActivities();

	public int getNumberOfEndActivities();

	/**
	 * This method considers graphs equal when they are the same and the
	 * activities were added in the same order.
	 * 
	 * @param other
	 * @return
	 */
	public boolean equals(DirectlyFollowsModel other);

	public int hashCode();

	public DirectlyFollowsModel clone();
}
